#!/usr/bin/python

#/* Copyright 2016 Research Foundation State University of New York   */

#/* This file is part of QUB Express.                                      */

#/* QUB Express is free software; you can redistribute it and/or modify    */
#/* it under the terms of the GNU General Public License as published by   */
#/* the Free Software Foundation, either version 3 of the License, or      */
#/* (at your option) any later version.                                    */

#/* QUB Express is distributed in the hope that it will be useful,         */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
#/* GNU General Public License for more details.                           */

#/* You should have received a copy of the GNU General Public License,     */
#/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
#/* <http://www.gnu.org/licenses/>.                                        */

import collections
import itertools
from numpy import *
import scipy
import scipy.linalg
import scipy.optimize
import sys
import traceback
import time
from qubx.fast.fast_utils import *
from qubx.model_constraints import *

def ApplyDeadTimeHJC(classes, durations, tdead):
    """ApplyDeadTime(classes, durations, tdead)

    classes:      list of each event's class index
    durations:    list of each event's duration, in milliseconds
    tdead:        length of shortest perceptible event
    
    Returns: (classes, durations) with
                 * events shorter than tdead joined to the prior event
                 * remaining events NOT shortened by tdead
"""
    # copy the arrays and modify them in-place
    cl = array(classes, copy=True, dtype='int32')
    du = array(durations, copy=True, dtype='float32')
    tm = 0.0
    L = len(du)
    if L == 0:
        return cl, du
    # skip initial too-shorts
    first_alive = 0
    while first_alive < L:
        if (du[first_alive] < tdead) and (fulpdiff(du[first_alive], tdead) > MAX_ULP_DIFF):
            tm += du[first_alive]
            first_alive += 1
        else:
            # du[first_alive] += tm # uncomment to join them to the first long-enough event
            break
    i_rd, i_wr = first_alive, 0
    while i_rd < L:
        cls, tm = cl[i_rd], du[i_rd]
        if i_wr and ((cl[i_wr-1] == cls) or ((tm < tdead) and (fulpdiff(tm, tdead) > MAX_ULP_DIFF))):
            du[i_wr-1] += tm
        else:
            cl[i_wr] = cls
            du[i_wr] = tm # - tdead
            i_wr += 1
        i_rd += 1
    cl.resize((i_wr))
    du.resize((i_wr))
    return cl, du

import ctypes
try:
    hjc = ctypes.cdll.LoadLibrary('qubxhjc.dll')
except:
    try:
        hjc = ctypes.cdll.LoadLibrary('libqubxhjc.so')
    except:
        hjc = ctypes.cdll.LoadLibrary('@executable_path/../Frameworks/libqubxhjc.so')

hjc.qub_hjc_bursts_alloc.argtypes = []
hjc.qub_hjc_bursts_alloc.restype = c_void_p
hjc.qub_hjc_bursts_free.argtypes = (c_void_p,)
hjc.qub_hjc_bursts_free.restype = None
hjc.qub_hjc_bursts_append.argtypes = (c_void_p, c_float_p, c_int)
hjc.qub_hjc_bursts_append.restype = None
hjc.qub_hjc_likelihood.argtypes = (c_void_p, c_double_p, c_int, c_int, c_double, c_double, c_int)
hjc.qub_hjc_likelihood.restype = c_double

class Bursts(object):
    def __init__(self):
        self.obj = hjc.qub_hjc_bursts_alloc()
        self.numEvents = 0
    def __del__(self):
        hjc.qub_hjc_bursts_free(self.obj)
    def append(self, classes, durationsMS, tdMS):
        cc, dd = ApplyDeadTimeHJC(classes, durationsMS, tdMS)
        if cc[0] != 1:
            raise Exception('Burst must start with class 1 (open) event')
        if cc[-1] != 1:
            raise Exception('Burst must end with class 1 (open) event')
        if 0 == (len(cc) % 2):
            raise Exception('Burst must contain an odd number of events')
        if any(cc > 1):
            raise Exception('Burst must contain only class 0 (closed) and 1 (open) events')
        hjc.qub_hjc_bursts_append(self.obj, dd.ctypes.data_as(c_float_p), len(dd))
        self.numEvents += len(dd)

def likelihood(bursts, Q, clazz, tdMS, tcritMS=1000.0, useCHSvectors=True):
    Nstate = Q.shape[0]
    # put Q matrix in HJC order: all open states in upper left corner
    Nopen = 0
    source_state = []
    for i, c in enumerate(clazz):
        if c == 1:
            Nopen += 1
            source_state.append(i)
    for i, c in enumerate(clazz):
        if c == 0:
            source_state.append(i)
        elif c > 1:
            raise Exception('All states must have class 0 (closed) or 1 (open)')
    Q_ordered = Q.copy()
    for i in xrange(Nstate):
        for j in xrange(Nstate):
            Q_ordered[i,j] = Q[source_state[i], source_state[j]]
    return hjc.qub_hjc_likelihood(bursts.obj, Q_ordered.ctypes.data_as(c_double_p), Nstate, Nopen, tdMS, tcritMS, int(useCHSvectors))

    
TDEAD = 0.1 # the sampling rate since the simulator is perfect
MODELFILE = 'msl_test.qmf'
DATAFILE = 'msl_test.qsf'

if __name__ == '__main__':
    import sys
    import qubx.model
    import qubx.tree
    qubx.tree.CHOOSE_FLAVOR('numpy')

    if '--help' in sys.argv:
        print """usage: qubxhjc.py [<tdead=0 in ms> [<model>.qmf [<data>.qsf]]]
or     qubxhjc.py --verify"""
        sys.exit(0)
    if '--verify' in sys.argv:
        import subprocess
        p = subprocess.Popen(sys.argv[0] + " | diff qubxhjc_test.out -", shell=True)
        p.communicate()
        sys.exit(0)

    def QSFtoSignals(session, wanted_signals=None):
        raw_idls = [(i, str(dc['Name'].data), chan)
                    for i, dc, chan in itertools.izip(itertools.count(),
                                                      qubx.tree.children(session['DataChannels'], 'Channel'),
                                                      qubx.tree.children(session['Idealization'], 'Channel'))
                    if not chan.find('Segment').isNull]
        wanted_idls = [(nm, chan) for i, nm, chan in raw_idls if ((i==0) or (wanted_signals==None) or (nm in wanted_signals))]
        raw_segs = [(nm, [(seg['Classes'].storage.data[:,0],
                           seg['Durations'].storage.data[:,0],
                           seg['amp'].storage.data[:,0])
                          for seg in qubx.tree.children(chan, 'Segment')])
                    for nm, chan in wanted_idls]
        return [nm for nm, segs in raw_segs], [segs for nm, segs in raw_segs]
        # [name_0, ...],. [(classes_0_seg0, durations_0_seg0, amp_of_cls_0_seg0), ...seg1...]

    tdead = (len(sys.argv) > 1) and float(sys.argv[1]) or TDEAD
    print 'Tdead:', tdead

    model_path = (len(sys.argv) > 2) and sys.argv[2] or MODELFILE
    data_path = (len(sys.argv) > 3) and sys.argv[3] or DATAFILE

    model = qubx.model.OpenQubModel(model_path)
    print model
    rates = model.rates
    wanted_signals = set([rates.get(r, 'Ligand') for r in xrange(rates.size)] +
                         [rates.get(r, 'Voltage') for r in xrange(rates.size)])

    qsf = qubx.tree.Open(data_path, True)
    signal_names, signals = QSFtoSignals(qsf, wanted_signals)
    segments, multi_classes, stimulus_classes = ProcessSignals(tdead, signals)
    print 'Events:', sum(len(cl) for cl,du in segments)

    td_sec = tdead * 1e-3
    clazz = qubx.model.ModelToClazz(model)
    K0, K1, K2, L, V, Pres = qubx.model.ModelToRateMatricesP(model, signal_names)
    K = K012toK(K0, K1, K2)


### TODO: use hjc Bursts and likelihood
    
    
    t0 = time.time()
    LL = subi_ll(clazz, None, K0, K1, K2, L, V, Pres, td_sec, segments, multi_classes, stimulus_classes, printout=True)
    print 'Initial LL:', LL, '[', ('%.3f' % (time.time() - t0)), 'sec ]'

    if HAVE_LIB:
        t0 = time.time()
        LL = subi_ll_lib(clazz, None, K0, K1, K2, L, V, Pres, td_sec, segments, multi_classes, stimulus_classes, printout=True)
        print 'Compiled LL:', LL, '[', ('%.3f' % (time.time() - t0)), 'sec ]'

    constraints = model.constraints_kin.get_matrices_p(K0, K1, K2, L, V, Pres)
    print 'Constraints in:'
    print Asys
    print Bsys
    Asys, Bsys = reduce_constraints(Asys, Bsys)
    Acns, Bcns, Ainv, pars = linear_constraints(Asys, Bsys, K)
    print 'Constraints out:'
    print Acns
    print Bcns
    Ascal, Bscal, Asci, pars = start_at_ones(pars)
    if any(pars != 1.0):
        print 'Non-unit starting par!',pars
        print Ascal
        print Bscal
        print Asci

    msl_ll = HAVE_LIB and subi_ll_lib or subi_ll
    def msl_func(pars):
        P = parsToK(pars, Ascal, Asci, Bscal)
        K = ParsToK(P, Acns, Ainv, Bcns)
        new_K0, new_K1, new_K2 = KtoK012(K, K0)
        LL = msl_ll(clazz, None, new_K0, new_K1, new_K2, L, V, Pres, td_sec, segments, multi_classes, stimulus_classes)
        print LL,pars
        return - LL
    counter = itertools.count()
    def msl_iter(pars):
        print '---> %i <---'%counter.next(), pars
        P = parsToK(pars, Ascal, Asci, Bscal)
        K = parsToK(P, Acns, Ainv, Bcns)
        print 'K',K.T
        new_K0, new_K1, new_K2 = KtoK012(K, K0)
        qubx.model.K012ToModel(new_K0, new_K1, new_K2, model)
        print model.rates
        print '------------->    <---'
    
    pars, fopt, iter, funcalls, warnflag = \
        scipy.optimize.fmin(msl_func, pars, maxfun=2000,
                            xtol=.0001, ftol=.0001, maxiter=300,
                            full_output=1, disp=1, retall=0, callback=msl_iter)
    print 'final',
    msl_iter(pars)


# ----------------------
import struct
from functools import partial

# (c) 2010 Eric L. Frederich
#
# Python implementation of algorithms detailed here...
# from http://www.cygnus-software.com/papers/comparingfloats/comparingfloats.htm

def c_mem_cast(x, f=None, t=None):
    '''
    do a c-style memory cast

    In Python...

    x = 12.34
    y = c_mem_cast(x, 'd', 'l')

    ... should be equivilent to the following in c...

    double x = 12.34;
    long   y = *(long*)&x;
    '''
    return struct.unpack(t, struct.pack(f, x))[0]

dbl_to_lng = partial(c_mem_cast, f='d', t='l')
lng_to_dbl = partial(c_mem_cast, f='l', t='d')
flt_to_int = partial(c_mem_cast, f='f', t='i')
int_to_flt = partial(c_mem_cast, f='i', t='f')

def ulp_diff_maker(converter, negative_zero):
    '''
    Getting the ulp difference of floats and doubles is similar.
    Only difference if the offset and converter.
    '''
    def the_diff(a, b):

        # Make a integer lexicographically ordered as a twos-complement int
        ai = converter(a)
        if ai < 0:
            ai = negative_zero - ai

        # Make b integer lexicographically ordered as a twos-complement int
        bi = converter(b)
        if bi < 0:
            bi = negative_zero - bi

        return abs(ai - bi)

    return the_diff

# double ULP difference
dulpdiff = ulp_diff_maker(dbl_to_lng, 0x8000000000000000)
# float  ULP difference
fulpdiff = ulp_diff_maker(flt_to_int, 0x80000000        )

# default to double ULP difference
ulpdiff = dulpdiff
ulpdiff.__doc__ = '''
Get the number of doubles between two doubles.
'''

MAX_ULP_DIFF = 4
